from dataclasses import dataclass
from typing import Self, Literal
from ..types import cst


type ObisGroup = int | set[int]
type GroupLiteral = Literal['a', 'b', 'c', 'd', 'e', 'f']


@dataclass
class LNPattern:
    """pattern for use in get_filtered.
    value "x.x.x.x.x.x" where x is:
    0..255 - simple value
    a,b,c,d,e,f - for skip value in each group
    (y, z, ...) - set of simple values(y or z)
    ((y-z), ...) - set of simple values with range(from y to z)
    !() - as () but, set of exclude values
    example: "a.0.(1,2,3).(0-64).0.f"
    """
    __values: tuple[ObisGroup, ...]

    def __post_init__(self):
        if len(self.__values) != 6:
            raise ValueError(F"for {self.__class__.__name__} got values with length={len(self.__values)}, expected 6")

    @classmethod
    def parse(cls, value: str) -> Self:
        values: list[ObisGroup] = [-1, -1, -1, -1, -1, -1]
        for i, val in enumerate(value.split('.', maxsplit=5)):
            if (
                len(val) == 1
                and (ord(val) == 97+i)
            ):
                if val == 'b':
                    values[i] = set(range(65))
                else:
                    continue
            elif val.isdigit():
                values[i] = int(val)
                if not (0 <= values[i] <= 255):
                    raise ValueError(F"in {value=} got element {val=}, expected 0..255")
            elif val.startswith('(') and val.endswith(')'):
                el: set[int] = set()
                val = val.replace('(', "").replace(')', "")
                for j in val.split(","):
                    j = j.replace(" ", '')
                    match j.count('-'):
                        case 0:
                            el.add(cls.__simple_validate(j))
                        case 1:
                            start, end = j.split("-")
                            el.update(range(
                                cls.__simple_validate(start),
                                cls.__simple_validate(end)+1))
                        case err:
                            raise ValueError(F"got a lot of <-> in pattern: {value}, expected one")
                values[i] = el
            elif val.startswith('!(') and val.endswith(')'):
                el: set[int] = set(range(256))
                val = val.replace('!(', "").replace(')', "")
                for j in val.split(","):
                    j = j.replace(" ", '')
                    match j.count('-'):
                        case 0:
                            el.discard(cls.__simple_validate(j))
                        case 1:
                            start, end = j.split("-")
                            el.difference_update(range(
                                cls.__simple_validate(start),
                                cls.__simple_validate(end)+1))
                        case err:
                            raise ValueError(F"got a lot of <-> in pattern: {value}, expected one")
                if len(el) == 0:
                    raise ValueError(F"no one element in group: {chr(97+i)}")
                values[i] = el
            else:
                raise ValueError(F"got wrong symbol: {val} in pattern")
        return cls(tuple(values))

    @staticmethod
    def __simple_validate(value: str) -> int:
        if value.isdigit() and (0 <= (new := int(value)) <= 255):
            return new
        else:
            raise ValueError(F"got not valid element: {value} in pattern, expected 0..255")

    def get_update(self, group: GroupLiteral, value: str) -> Self:
        """get new instance with updated group value"""
        # todo: make this
        raise RuntimeError("no implement now")

    def __eq__(self, other: cst.LogicalName):
        for i, j in zip(self.__values, other):
            if i == j or (i == -1):
                continue
            elif isinstance(i, set) and j in i:
                continue
            else:
                return False
        return True

    def __repr__(self):
        return F"{self.__class__.__name__}(\"{".".join(map(lambda it: str(it) if isinstance(it, int) else str(tuple(it)), self.__values))}\")"


@dataclass
class LNPatterns:
    value: tuple[LNPattern, ...]

    def __iter__(self):
        return iter(self.value)


ABSTRACT = LNPattern.parse("0")
ELECTRICITY = LNPattern.parse("1")
HCA = LNPattern.parse("4")
THERMAL = LNPattern.parse("(5,6)")
GAS = LNPattern.parse("7")
WATER = LNPattern.parse("(8,9)")
OTHER_MEDIA = LNPattern.parse("15")


BILLING_PERIOD_VALUES_RESET_COUNTER_ENTRIES = LNPatterns((
    LNPattern.parse("0.b.0.1.(0,2,3,5).f"),
    LNPattern.parse("0.b.0.1.(1,4).255")))
PROGRAM_ENTRIES = LNPatterns((
    ACTIVE_FIRMWARE_IDENTIFIER := LNPattern.parse("0.b.0.2.0.255"),
    ACTIVE_FIRMWARE_VERSION := LNPattern.parse("0.b.0.2.1.255"),
    ACTIVE_FIRMWARE_SIGNATURE := LNPattern.parse("0.b.0.2.8.255")
))
TIME_ENTRIES = LNPattern.parse("0.b.0.9.(1,2).255")
CLOCK_OBJECTS = LNPatterns((
    CLOCK := LNPattern.parse("0.b.1.0.e.255"),
    UNIX_CLOCK := LNPattern.parse("0.b.1.1.e.255"),
    MICROSECONDS_CLOCK := LNPattern.parse("0.b.1.2.e.255"),
    MINUTES_CLOCK := LNPattern.parse("0.b.1.3.e.255"),
    HOURS_CLOCK := LNPattern.parse("0.b.1.4.e.255"),
    DAYS_CLOCK := LNPattern.parse("0.b.1.5.e.255"),
    WEEKS_CLOCK := LNPattern.parse("0.b.1.6.e.255")
))
TARIFFICATION_SCRIPT_TABLE = LNPattern.parse("0.b.10.0.100.255")
PUSH_SCRIPT_TABLE = LNPattern.parse("0.b.10.0.108.255")
SINGLE_ACTION_SCHEDULE = LNPatterns((
    END_OF_BILLING_PERIOD_SINGLE_ACTION_SCHEDULE := LNPattern.parse("0.b.15.0.0.255"),
    DISCONNECT_CONTROL_SINGLE_ACTION_SCHEDULE := LNPattern.parse("0.b.15.0.1.255"),
    IMAGE_ACTIVATION_SINGLE_ACTION_SCHEDULE := LNPattern.parse("0.b.15.0.2.255"),
    OUTPUT_CONTROL_SINGLE_ACTION_SCHEDULE := LNPattern.parse("0.b.15.0.3.255"),
    PUSH_SINGLE_ACTION_SCHEDULE := LNPattern.parse("0.b.15.0.4.255"),
    LOAD_PROFILE_CONTROL_SINGLE_ACTION_SCHEDULE := LNPattern.parse("0.b.15.0.5.255"),
    M_BUS_PROFILE_CONTROL_SINGLE_ACTION_SCHEDULE := LNPattern.parse("0.b.15.0.6.255"),
    FUNCTION_CONTROL_SINGLE_ACTION_SCHEDULE := LNPattern.parse("0.b.15.0.7.255")
))
ACTIVITY_CALENDAR = LNPattern.parse("0.b.13.0.e.255")
ASSOCIATION = LNPattern.parse("0.0.40.0.e.255")  # 6_2_33
NON_CURRENT_ASSOCIATION = LNPattern.parse("0.0.40.0.(1-255).255")  # MY
SAP_ASSIGNMENT = LNPattern.parse("0.0.41.0.0.255")  # 6.2.34
COSEM_logical_device_name = LNPattern.parse("0.0.42.0.0.255")  # 6.2.35
INFORMATION_SECURITY_RELATED = LNPatterns((
    LNPattern.parse("0.0.43.(0,2).e.255"),
    INVOCATION_COUNTER := LNPattern.parse("0.b.43.1.e.255")  # 6.2.36
))
IMAGE_TRANSFER = LNPattern.parse("0.b.44.0.e.255")  # 6.2.37
FUNCTION_CONTROL = LNPattern.parse("0.b.44.1.e.255")  # 6.2.38
COMMUNICATION_PORT_PROTECTION = LNPattern.parse("0.b.44.2.e.255")  # 6.2.39
UTILITY_TABLE = LNPattern.parse("0.b.65.(0-63).e.255")  # 6.2.40
COMPACT_DATA = LNPattern.parse("0.b.66.0.e.255")  # 6.2.41
DEVICE_ID = LNPattern.parse("0.b.96.1.(0,1,2,3,4,5,6,7,8,9,255).255")  # 6.2.42
METERING_POINT_ID = LNPattern.parse("0.b.96.1.10.255")  # 6.2.43
PARAMETER_CHANGES_CALIBRATION_AND_ACCESS = LNPattern.parse("0.b.96.2.e.f")  # 6.2.44
INPUT_OUTPUT_CONTROL_SIGNALS = LNPattern.parse("0.b.96.3.(0-4).f")  # 6.2.45
DISCONNECT_CONTROL = LNPattern.parse("0.b.96.3.10.f")  # 6.2.46
ARBITRATOR = LNPattern.parse("0.b.96.3.(20-29).f")  # 6.2.47
INTERNAL_CONTROL_SIGNALS = LNPattern.parse("0.b.96.4.(0-4).f")  # 6.2.48
INTERNAL_OPERATING_STATUS = LNPattern.parse("0.b.96.5.(0-4).f")  # 6.2.49
BATTERY_ENTRIES = LNPattern.parse("0.b.96.6.(0,1,2,3,4,5,6,10,11).f")  # 6.2.50
POWER_FAILURE_MONITORING = LNPattern.parse("0.b.96.7.(0-21).f")  # 6.2.51
OPERATING_TIME = LNPattern.parse("0.b.96.8.(0-63).f")
ENVIRONMENT_RELATED_PARAMETERS = LNPattern.parse("0.b.96.9.(0-2).f")
STATUS_REGISTER = LNPattern.parse("0.b.96.10.(1-10).f")
EVENT_CODE = LNPattern.parse("0.b.96.11.(0-99).f")
COMMUNICATION_PORT_LOG_PARAMETERS = LNPattern.parse("0.b.96.12.(0-6).f")
CONSUMER_MESSAGES = LNPattern.parse("0.b.96.13.(0,1).f")
CURRENTLY_ACTIVE_TARIFF = LNPattern.parse("0.b.96.14.(0-15).f")
EVENT_COUNTER = LNPattern.parse("0.b.96.15.(0-99).f")
PROFILE_ENTRY_DIGITAL_SIGNATURE = LNPattern.parse("0.b.96.16.(0-9).f")
PROFILE_ENTRY_COUNTER = LNPattern.parse("0.b.96.17.(0-127).f")
METER_TAMPER_EVENT_RELATED = LNPattern.parse("0.b.96.20.(0-34).f")
MANUFACTURER_SPECIFIC_ABSTRACT = LNPattern.parse("0.b.96.(50-99).e.f")

GENERAL_AND_SERVICE_ENTRY = LNPatterns((
    *BILLING_PERIOD_VALUES_RESET_COUNTER_ENTRIES,
    *PROGRAM_ENTRIES,
    TIME_ENTRIES,
    DEVICE_ID,
    PARAMETER_CHANGES_CALIBRATION_AND_ACCESS,
    INPUT_OUTPUT_CONTROL_SIGNALS,
    INTERNAL_CONTROL_SIGNALS,
    INTERNAL_OPERATING_STATUS,
    BATTERY_ENTRIES,
    POWER_FAILURE_MONITORING,
    OPERATING_TIME,
    ENVIRONMENT_RELATED_PARAMETERS,
    STATUS_REGISTER,
    EVENT_CODE,
    COMMUNICATION_PORT_LOG_PARAMETERS,
    CONSUMER_MESSAGES,
    CURRENTLY_ACTIVE_TARIFF,
    EVENT_COUNTER,
    PROFILE_ENTRY_DIGITAL_SIGNATURE,
    PROFILE_ENTRY_COUNTER,
    METER_TAMPER_EVENT_RELATED,
    MANUFACTURER_SPECIFIC_ABSTRACT))
"""DLMS UA 1000-1 Ed. 14 7.4.1"""

LIMITER = LNPattern.parse("0.b.17.0.e.255")  # 6.2.15
ALARM_REGISTER = LNPattern.parse("0.b.97.98.(0-9).255")  # 6.2.64
COUNTRY_SPECIFIC_IDENTIFIERS = LNPattern.parse("0.b.94.d.e.f")  # 7.3.4.3
ALARM_REGISTER_FILTER = LNPattern.parse("0.b.97.98.(10-19).255")  # 6.2.64
ALARM_REGISTER_DESCRIPTOR = LNPattern.parse("0.b.97.98.(20-29).255")  # 6.2.64
ALARM_REGISTER_PROFILE = LNPattern.parse("0.b.97.98.255.255")  # 6.2.64
ALARM_REGISTER_TABLE = LNPattern.parse("0.b.97.98.255.255")  # 6.2.64
ALARM_REGISTER_FILTER_DESCRIPTOR = LNPatterns((ALARM_REGISTER, ALARM_REGISTER_FILTER, ALARM_REGISTER_DESCRIPTOR, ALARM_REGISTER_PROFILE))
# electricity
ID_NUMBERS_ELECTRICITY = LNPattern.parse("1.b.0.0.(0-9).255")
ELECTRIC_PROGRAM_ENTRIES = LNPattern.parse("1.b.0.2.e.255")
OUTPUT_PULSE_VALUES_OR_CONSTANTS = LNPattern.parse("1.0.0.3.(0-9).255")
RATIOS = LNPattern.parse("1.0.0.4.(0-7).255")
RECORDING_INTERVAL = LNPattern.parse("1.0.0.8.(4,5).255")
OTHER_ELECTRICITY_RELATED_GENERAL_PURPOSE = LNPattern.parse("1.b.0.(2,3,4,6,7,8,9,10).e.255")
# my special
COUNTRY_SPECIFIC = LNPattern.parse("a.b.94.d.e.f")  # 7.2.4 Table 54

# 7.5.2.1
INSTANTANEOUS_VALUE_EL_REL = LNPattern.parse("1.b.!(0,93,94,96,97,98,99).7.e.f")
ELECTRICITY_RELATED = LNPatterns((
    INSTANTANEOUS_VALUE_EL_REL,
    # todo: more other LNPatterns
))
"""7.5.2.1 Processing of measurement values Table 66"""



import unittest
from src.DLMS_SPODES.types import cdt, cst, ut
from src.DLMS_SPODES.cosem_interface_classes import collection, overview


class TestType(unittest.TestCase):

    def test_Data(self):
        obj = collection.impl.profile_generic.SPODES3DisplayReadout("0.0.21.0.1.255")
        print(obj)


import unittest
from src.DLMS_SPODES.cosem_interface_classes import a_parameter


class TestType(unittest.TestCase):

    def test_clock_time(self):
        value = a_parameter.Time()
        print(value.value)



import os
from pydantic import BaseModel, Field
import tomllib
from enum import Enum


class _ReportStruct(BaseModel):
    """
    Patterns. Use for function get_obj_report. Only for struct. Generate custom struct report instead standart. Where %nxx: n - name of struct element,
    xx - index of struct element; %vxx: v - value of struct element, xx - index of struct element
    """
    DayProfileAction: str = "%n00: %v00 - Тариф: %v02"


class _Report(BaseModel):
    empty: str = "--"
    empty_unit: str = "??"
    scaler_format: str = "{:.3f}"
    struct: _ReportStruct = Field(default_factory=_ReportStruct)


class _Collection(BaseModel):
    path: str = "./Types/"


class Settings(BaseModel):
    collection: _Collection = Field(default_factory=_Collection)
    report: _Report = Field(default_factory=_Report)


if not os.path.isfile(path := ".//config.toml"):
    path = F"{os.path.dirname(__file__)}{path}"
elif os.path.isfile(path):
    with open(path, "rb") as f:
        toml_data = tomllib.load(f)
        data = toml_data.get("VIEW", {})
        print(f"Find configuration <config.toml> with path: {f}")
        settings = Settings(**data)
else:
    print("NOT FIND CONFIGURATION: <config.toml>")
    toml_data = {}
    settings = Settings()


# remove in future


def version():
    return "0.2.0"


class Language(Enum):
    ENGLISH = 'English'
    RUSSIAN = 'Russian'


__current_language = Language.RUSSIAN


def set_current_language(value: str):
    global __current_language
    __current_language = Language(value)


def get_current_language() -> Language:
    return __current_language



from ...types import common_data_types as cdt
from ...config_parser import get_values

base = get_values("DLMS", "Conformance")


# TODO: join with cdt.FlagMixin
class Conformance(cdt.BitString):
    ELEMENTS = ("reserved-zero",
                "general-protection",
                "general-block-transfer",
                "read",
                "write",
                "unconfirmed-write",
                "reserved-six",
                "reserved-seven",
                "attribute0-supported-with-set",
                "priority-mgmt-supported",
                "attribute0-supported-with-get",
                "block-transfer-with-get-or-read",
                "block-transfer-with-set-or-write",
                "block-transfer-with-action",
                "multiple-references",
                "information-report",
                "data-notification",
                "access",
                "parameterized-access",
                "get",
                "set",
                "selective-access",
                "event-notification",
                "action")
    if base is not None:
        ELEMENTS = tuple(base[el] for el in ELEMENTS)
    default = '011111111111111111111111'  # zero only 1 bit

    def __init__(self, value: bytes | bytearray | str | int | cdt.BitString = None):
        super(Conformance, self).__init__(value)
        if self.ELEMENTS is not None and len(self) != len(self.ELEMENTS):
            raise ValueError(F'For {self.__class__.__name__} get {len(self)} bits, expected {len(self.ELEMENTS)}')

    def __len__(self):
        return 24

    def from_bytes(self, value: bytes) -> bytes:
        length, pdu = cdt.get_length_and_pdu(value[1:])
        if length != len(self):
            raise ValueError(F'Got {length=}, expected {len(self)}')
        match value[:1]:
            case self.TAG if len(self) <= len(pdu) * 8: return pdu[:3]
            case self.TAG:                              raise ValueError(F'Got pdu length:{len(pdu)}, expected at least {len(self) >> 3}')
            case _ as error:                            raise TypeError(F'Expected {self.NAME} type, got {cdt.get_common_data_type_from(error).NAME}')

    def from_str(self, value: str) -> bytes:
        value = value + '0' * ((8 - len(self)) % 8)
        list_ = [value[count:(count + 8)] for count in range(0, len(self), 8)]
        value = b''
        for byte in list_:
            value += int(byte, base=2).to_bytes(1, byteorder='little')
        return value

    def from_int(self, value: int) -> bytes:
        if value < 0:
            raise ValueError
        res = 0
        start_bit = 2 ** (len(self) - 1)
        for i in range(len(self)):
            if value & (1 << i):
                res += start_bit >> i
        return res.to_bytes(len(self) // 8, byteorder='big')

    def from_bytearray(self, value: bytearray) -> bytes:
        return bytes(value)

    @classmethod
    def get_values(cls) -> list[str]:
        """ TODO: """
        return cls.ELEMENTS

    def validate_from(self, value: str, cursor_position: int) -> tuple[str, int]:
        """ return validated value and cursor position. TODO: copypast FlagMixin """
        type(self)(value=value.zfill(len(self)))
        return value, cursor_position

    @property
    def general_protection(self) -> int:
        return tuple(self)[1]

    @property
    def general_block_transfer(self) -> int:
        return tuple(self)[2]

    @property
    def selective_access(self) -> int:
        return tuple(self)[21]



from .__class_init__ import *
from ..types import choices


class ServiceId(cdt.Enum, elements=(1, 2)):
    """defines which action to be applied to the referenced object."""


class ActionSpecification(cdt.Structure):
    """ Specifies the different scripts, i.e. the lists of actions. The first attribute (logical_name) has index 1, the first specific method has
    index 1 as well. NOTE The action_specification is limited to activate methods that do not produce any response (from the server to the client). """
    service_id: ServiceId
    class_id: cdt.LongUnsigned
    logical_name: cst.LogicalName
    index: cdt.Integer
    parameter: choices.common_dt


class Actions(cdt.Array):
    """ Specifies the list of action specification """
    TYPE = ActionSpecification


class Script(cdt.Structure):
    """ Specifies the different scripts. The script_identifier 0 is reserved. If specified with an execute method, it results in a null script (no actions to perform)"""
    script_identifier: cdt.LongUnsigned
    actions: Actions


class Scripts(cdt.Array):
    """ Specifies the lists of actions """
    TYPE = Script
    __get_item__: Script

    def new_element(self) -> Script:
        """return default Script with vacant script_identifier"""
        # todo: make common function by ID element
        el: Script
        day_ids: list[int] = [int(el.script_identifier) for el in self.values]
        for i in range(0xffff):
            if i not in day_ids:
                return Script((i, None))
        raise ValueError(F"in {self} all <script_identifier> is busy")


class ScriptTable(ic.COSEMInterfaceClasses):
    """Script table"""
    CLASS_ID = ClassID.SCRIPT_TABLE
    VERSION = Version.V0
    A_ELEMENTS = ic.ICAElement("scripts", Scripts),
    M_ELEMENTS = ic.ICMElement("execute", cdt.LongUnsigned),

    def characteristics_init(self):
        self.set_attr(2, None)
        self._cbs_attr_post_init.update({2: self.__set_script_identifier_cbs})

    @property
    def scripts(self) -> Scripts:
        return self.get_attr(2)

    @property
    def execute(self) -> cdt.LongUnsigned:
        return self.get_meth(1)

    def __set_script_identifier_cbs(self):
        pass
        # try:
        #     indexes: Callable = self.entries.get_indexes
        #     self.enable_disable.firstIndexA.set_callback(indexes)
        #     self.enable_disable.firstIndexB.set_callback(indexes)
        #     self.enable_disable.lastIndexA.set_callback(indexes)
        #     self.enable_disable.lastIndexB.set_callback(indexes)
        #     self.insert.index.set_callback(indexes)
        #     self.delete.firstIndex.set_callback(indexes)
        #     self.delete.lastIndex.set_callback(indexes)
        #     print('set delete')
        # except KeyError:  # At init time
        #     print('set delete NO:')


if __name__ == '__main__':
    a = b'\x01\x04\x02\x02\x12\x00\x01\x01\x01\x02\x05\x16\x01\x12\x00F\t\x06\x00\x00`\x03\n\xff\x0f\x04\x16\x00\x02\x02\x12\x00\x02\x01\x01\x02\x05\x16\x02\x12\x00F\t\x06\x00\x00`\x03\n\xff\x0f\x01\x03\x00\x02\x02\x12\x00\x03\x01\x01\x02\x05\x16\x01\x12\x00F\t\x06\x00\x00`\x03\n\xff\x0f\x04\x16\x01\x02\x02\x12\x00\x04\x01\x01\x02\x05\x16\x02\x12\x00F\t\x06\x00\x00`\x03\n\xff\x0f\x01\x03\x01'
    b = Scripts(a)
    a = ScriptTable('0.0.10.0.0.255')
    pass
    print(a)



import logging

from ..__class_init__ import *
from ...config_parser import get_message


class Status(cdt.Enum, elements=tuple(range(6))):
    """ Indicates the registration status of the  modem. """


class CSAttachment(cdt.Enum, elements=(0, 1, 2)):
    """ Indicates the current circuit switched status."""


class PSStatus(cdt.Enum, elements=tuple(range(5))):
    """ Indicates the packet switched status of the modem. """


class SignalQuality(cdt.IntegerEnum, cdt.Unsigned):
    """for string report"""
    def __init_subclass__(cls, **kwargs):
        """not need"""

    def get_report(self) -> cdt.Report:
        val = int(self)
        if val == 0:
            return cdt.Report(get_message(F"({val}) –113 dBm $or$ $less$(0)"))
        elif val == 1:
            return cdt.Report(F"({val}) –111 dBm")
        elif val < 31:
            return cdt.Report(get_message(F"({val}) {-109+(val-2)*2} dBm"))
        elif val == 31:
            return cdt.Report(get_message(F"({val}) –51 dBm $or$ $greater$"))
        elif val == 99:
            return cdt.Report(get_message(F"({val}) $not_known_or_not_detectable$"))
        else:
            return cdt.Report(
                msg=F"({val})",
                log=cdt.Log(logging.WARN, "unknown value"))


class CellInfoType(cdt.Structure):
    """ Params of element """
    cell_ID: cdt.LongUnsigned
    location_ID: cdt.LongUnsigned
    signal_quality: SignalQuality
    ber: cdt.Unsigned


class AdjacentCellInfo(cdt.Structure):
    cell_ID: cdt.LongUnsigned
    signal_quality: SignalQuality


class AdjacentCells(cdt.Array):
    TYPE = AdjacentCellInfo


class GSMDiagnostic(ic.COSEMInterfaceClasses):
    """ The GSM/GPRS network is undergoing constant changes in terms of registration status, signal quality etc. It is necessary to monitor and log the relevant parameters in order
     to obtain diagnostic information that allows identifying communication problems in the network. An instance of the 'GSM diagnostic' class stores parameters of the GSM/GPRS
     network necessary for analysing the operation of the network."""
    CLASS_ID = ClassID.GSM_DIAGNOSTIC
    VERSION = Version.V0
    A_ELEMENTS = (ic.ICAElement("operator", cdt.VisibleString, classifier=ic.Classifier.DYNAMIC),
                  ic.ICAElement("status", Status, 0, 255, 0, classifier=ic.Classifier.DYNAMIC),
                  ic.ICAElement("cs_attachment", CSAttachment, 0, 255, 0, classifier=ic.Classifier.DYNAMIC),
                  ic.ICAElement("ps_status", PSStatus, 0, 255, 0, classifier=ic.Classifier.DYNAMIC),
                  ic.ICAElement("cell_info", CellInfoType, classifier=ic.Classifier.DYNAMIC),
                  ic.ICAElement("adjacent_cell", AdjacentCells, classifier=ic.Classifier.DYNAMIC),
                  ic.ICAElement("capture_time", cdt.DateTime, classifier=ic.Classifier.DYNAMIC))

    def characteristics_init(self):
        """nothing do it"""

    @property
    def operator(self) -> cdt.VisibleString:
        return self.get_attr(2)

    @property
    def status(self) -> Status:
        return self.get_attr(3)

    @property
    def cs_attachment(self) -> CSAttachment:
        return self.get_attr(4)

    @property
    def ps_status(self) -> PSStatus:
        return self.get_attr(5)

    @property
    def cell_info(self) -> CellInfoType:
        return self.get_attr(6)

    @property
    def adjacent_cell(self) -> AdjacentCells:
        return self.get_attr(7)

    @property
    def capture_time(self) -> cdt.DateTime:
        return self.get_attr(8)



import unittest
from src.DLMS_SPODES.cosem_interface_classes import overview


class TestType(unittest.TestCase):
    def test_ClassID(self):
        value = overview.ClassID.DATA
        print(value)



from ..arbitrator import Arbitrator
from ... import settings


match settings.get_current_language():
    case settings.Language.ENGLISH:        from ...Values.EN import actors
    case settings.Language.RUSSIAN:        from ...Values.RU import actors


class SPODES3Arbitrator(Arbitrator):
    """Cosem3 7.3.18"""
    actors = (actors.MANUAL,
              actors.LOCAL_1,
              actors.LOCAL_2,
              actors.LOCAL_3,
              actors.LOCAL_4,
              actors.LOCAL_5,
              actors.LOCAL_6,
              actors.LOCAL_7)



from ..__class_init__ import *
from . import ver0, ver1
from ...types.implementations import structs, arrays


class AssociationLN(ver1.AssociationLN):
    """5.4.7 Association LN"""
    VERSION = Version.V2
    A_ELEMENTS = (
        ver1.AssociationLN.get_attr_element(2),  # <object_list>
        ver0.AssociationLN.get_attr_element(3),  # associated_partners_id
        ver0.AssociationLN.get_attr_element(4),  # application_context_name
        ver0.AssociationLN.get_attr_element(5),  # xDLMS_context_info
        ver0.AssociationLN.get_attr_element(6),  # authentication_mechanism_name
        ver1.AssociationLN.get_attr_element(7),  # secret
        ver0.AssociationLN.get_attr_element(8),  # association_status
        ver1.AssociationLN.get_attr_element(9),  # security_setup_reference
        ic.ICAElement("user_list", arrays.UserList),
        ic.ICAElement("current_user", structs.UserListEntry),
    )
    M_ELEMENTS = (
        ver0.AssociationLN.get_meth_element(1),
        ver0.AssociationLN.get_meth_element(2),
        ver1.AssociationLN.get_meth_element(3),  # add_object
        ver1.AssociationLN.get_meth_element(4),  # remove_object
        ic.ICMElement("add_user", structs.UserListEntry),
        ic.ICMElement("remove_user", structs.UserListEntry),
    )
    user_list: arrays.UserList
    current_user: structs.UserListEntry

    def characteristics_init(self):
        super(AssociationLN, self).characteristics_init()
        # TODO: more 2 attribute
        # TODO: more 2 methods



from src.DLMS_SPODES.cosem_interface_classes import collection, overview
from src.DLMS_SPODES.types import cdt, ut, cst


col = collection.Collection(
    id_=collection.ID(
        man=b'KPZ',
        f_id=collection.ParameterValue(
            par=b'',
            value=cdt.OctetString("4d324d5f33").encoding),
        f_ver=collection.ParameterValue(
            par=b'',
            value=cdt.OctetString(bytearray(b'1.3.0')).encoding)
    )
)
col.spec_map = col.get_spec()
print(col)
ass = col.add(
    class_id=overview.ClassID.ASSOCIATION_LN,
    version=overview.Version.V1,
    logical_name=cst.LogicalName.from_obis("0.0.40.0.1.255"))


import unittest
from src.DLMS_SPODES.config_parser import get_message, get_values


class TestType(unittest.TestCase):
    def test_get_message(self):
        self.assertEqual("hello world", get_message('hello world'), "simple test")
        print(get_message("$or$"))
        self.assertEqual(get_message("$or$"), "или", "check translate to rus")
        self.assertEqual(get_message("$$or$$"), "$or$", "check translate to rus")
        print(get_message("–113 dBm $or$ $less$(0)"))

    def test_firmwares(self):
        from src.DLMS_SPODES.firmwares import get_firmware

        firmwares = get_firmware(b"KPZ")
        print(firmwares)



import unittest
from src.DLMS_SPODES.types import cdt, cst, ut
from src.DLMS_SPODES.cosem_interface_classes import collection, overview
from src.DLMS_SPODES import cosem_interface_classes
from src.DLMS_SPODES.exceptions import NeedUpdate, NoObject
from src.DLMS_SPODES.cosem_interface_classes import reports
from .init_collection import col


class TestType(unittest.TestCase):

    def test_report(self):
        # col = collection.get_collection(
        #     manufacturer=b"KPZ",
        #     server_type=cdt.OctetString("4d324d5f33"),
        #     server_ver=AppVersion.from_str("1.4.15"))
        print(col)
        obj = col.get_object("0.0.1.0.0.255")
        obj.set_attr(2, "01.09.2023")
        obj.set_attr(3, 180)
        rep = reports.get_obj_report(
            obj=obj,
            attr_index_par=(2, 3))
        print(rep)
        obj = col.get_object("0.0.96.1.0.255")
        obj.set_attr(2, "30313233")
        rep = reports.get_obj_report(
            obj=obj,
            attr_index_par=(2,))
        print(rep)
        obj = col.get_object("0.0.13.0.0.255")
        obj.set_attr(5, [
            [0, [("10:00", "1.1.1.1.1.1", 1)]], [1, [("12:00", "1.1.1.1.1.1", 2), ("14:00", "1.1.1.1.1.1", 3)]]])
        rep = reports.get_obj_report(
            obj=obj,
            attr_index_par=(
                3,
                4,
                (5, None)
            ))
        print(rep)



class OBIS(bytes):
    """bytes[6]"""

    def validate(self):
        if (length := len(self)) != 6:
            return ValueError(f"got {length=}, expected 6")

    def __str__(self):
        return F"{".".join(map(str, self[:6]))}"


class AssociationLN(OBIS):
    def __new__(cls, e: int):
        return super().__new__(cls, (0, 0, 40, 0, e, 255))


CURRENT_ASSOCIATION = OBIS((0, 0, 40, 0, 0, 255))
PUBLIC_ASSOCIATION = OBIS((0, 0, 40, 0, 1, 255))
LDN = OBIS((0, 0, 42, 0, 0, 255))


# SPODES_3
SPODES3_DISPLAY_MODE = OBIS((0, 0, 96, 4, 1, 255))



"""DLMS UA 1000-1 Ed 14"""
from typing import Type
from ..__class_init__ import *
from ...types.implementations import long_unsigneds, integers


BUFFER = 2
CAPTURE_OBJECTS = 3
CAPTURE_PERIOD = 4
SORT_METHOD = 5
SORT_OBJECT = 6
ENTRIES_IN_USE = 7
PROFILE_ENTRIES = 8


class ObjectDefinition(cdt.Structure):
    """ObjectDefinition"""
    logical_name: cst.LogicalName
    class_id: long_unsigneds.ClassId
    attribute_index: cdt.Integer

    def __hash__(self):
        return hash((self.logical_name.contents, self.class_id.contents, self.attribute_index.contents))


class CaptureObjects(cdt.Array):
    """ Specifies the list of capture objects """
    TYPE = ObjectDefinition


class SortMethod(cdt.Enum, elements=(1, 2, 3, 4, 5, 6)):
    """sort_method"""


class ProfileGeneric(ic.COSEMInterfaceClasses):
    """5.3.1 Profile generic"""
    CLASS_ID = ClassID.PROFILE_GENERIC
    VERSION = Version.V0
    # todo: remove down following
    scaler_profile_key: bytes | None = None
    """ obis of scaler profile for this profile if need """
    buffer_capture_objects: CaptureObjects
    range_descriptor: Type[cdt.Structure] = None
    attr_descriptor_with_selection: Type[ut.CosemAttributeDescriptorWithSelection] = None

    A_ELEMENTS = (
        ic.ICAElement("buffer", cdt.Array, classifier=ic.Classifier.DYNAMIC),
        ic.ICAElement("capture_objects", CaptureObjects),
        ic.ICAElement("capture_period", cdt.DoubleLongUnsigned),  # 5.3.1 Profile generic. capture_period
        ic.ICAElement("sort_method", SortMethod),
        ic.ICAElement("sort_object", ObjectDefinition),
        ic.ICAElement("entries_in_use", cdt.DoubleLongUnsigned, 0, default=0, classifier=ic.Classifier.DYNAMIC),
        ic.ICAElement("profile_entries", cdt.DoubleLongUnsigned, 1, default=1)
    )
    M_ELEMENTS = (
        ic.ICMElement("reset", integers.Only0),
        ic.ICMElement("capture", integers.Only0),
        # more 2 elements
    )

    def characteristics_init(self):
        self.set_attr(BUFFER, None)

        # todo remove it
        self.buffer.register_cb_preset(lambda _: self.__create_buffer_struct_type())  # value not used for creating struct type

        self._cbs_attr_post_init.update({CAPTURE_OBJECTS: self.__create_buffer_struct_type})

        self.buffer_capture_objects = self.capture_objects
        """ objects for buffer. Change with access_selection """

    @property
    def buffer(self) -> cdt.Array:
        return self.get_attr(2)

    @property
    def capture_objects(self) -> CaptureObjects:
        return self.get_attr(3)

    @property
    def capture_period(self) -> cdt.DoubleLongUnsigned:
        return self.get_attr(4)

    @property
    def sort_method(self) -> SortMethod:
        return self.get_attr(5)

    @property
    def sort_object(self) -> ObjectDefinition:
        return self.get_attr(6)

    @property
    def entries_in_use(self) -> cdt.DoubleLongUnsigned:
        return self.get_attr(7)

    @property
    def profile_entries(self) -> cdt.DoubleLongUnsigned:
        return self.get_attr(8)

    @property
    def reset(self) -> integers.Only0:
        return self.get_meth(1)

    @property
    def capture(self) -> integers.Only0:
        return self.get_meth(2)

    def __create_buffer_struct_type(self):
        """ TODO: more refactoring !!! """
        # rename ObjectDefinition's and adding object if it absense in collection
        el_value: ObjectDefinition
        if self.capture_objects is None:
            raise ValueError(F"{self}: create buffer Struct type, not set <capture_object> attribute. Need initiate <capture_objects> before")
        for el_value in self.capture_objects:
            self.collection.add_if_missing(
                class_id=ut.CosemClassId(el_value.class_id.contents),
                version=None,
                logical_name=el_value.logical_name)
            el_value.set_name(self.collection.get_name_and_type(el_value)[0][-1])
        self.buffer_capture_objects = self.capture_objects
        self.buffer.set_type(self.capture_objects.__class__)



from typing_extensions import deprecated
from dataclasses import dataclass
from struct import Struct, pack, unpack_from
from typing import Self, Optional, cast
import re
from .. import exceptions as exc
from .obis import OBIS

_pattern = re.compile("((?:\d{1,3}\.){5}\d{1,3})(?::(m?\d{1,3}))?")
Index = Struct("?B")


@dataclass(frozen=True)
class Parameter:
    """
     Parameter ::= SEQUENCE
     {
        ln          Cosem-Object-Instance-Id
        descriptor  OPTIONAL
    }

    index       Unsigned8
    nest_index  Unsigned16
    piece       Unsigned8

    descriptor :: = CHOICE
    {
        attribute   [0] IMPLICIT Desc
        method      [1] IMPLICIT Desc
    }

    Desc ::= SEQUENCE
    {
        index
        SEQUENCE (SIZE(255)) OF nest_index OPTIONAL
        piece OPTIONAL
    }
    """
    value: bytes

    def __bytes__(self):
        return self.value

    @classmethod
    def parse(cls, value: str) -> Self:
        """create from string. Only LN, attr/meth type ddd.ddd.ddd.ddd.ddd.ddd:aaa, ex.: 0.0.1.0.0.255 """
        if (res := _pattern.fullmatch(value)) is None:
            raise ValueError(F"in {cls.__name__}.parse got wrong :{value:}")
        else:
            groups = iter(res.groups())
            ret = bytes(map(int, next(groups).split(".")))
            if (a := next(groups)) is not None:
                if a.startswith('m'):
                    a = a[1:]
                    g1 = 256
                else:
                    g1 = 0
                ret += (g1 + int(a)).to_bytes(2)
            return cls(ret)

    def __eq__(self, other) -> bool:
        if isinstance(other, Parameter):
            return cast("bool", self.value == other.value)
        return NotImplemented

    def __lt__(self, other: Self):
        """comparing for sort method"""
        if len(self.value) > len(other.value):
            return True
        else:
            return False

    def __str__(self):
        if (l := len(self.value)) < 6:
            return "No valid"
        elif l == 7:
            return "No valid Index"
        else:
            res = F"{".".join(map(str, self.value[:6]))}"
        if l > 6:
            res += F":{"m" if self.is_method() else ""}{self.i}"
        if l > 8:
            res += F" {"/".join(map(str, self.elements()))}"
        if self.has_piece():
            res += F"p{self.piece}"
        return res

    def validate(self):
        if (length := len(self.value)) < 6:
            raise exc.DLMSException(F"Parameter got {length=}, expected at least 6")
        if length == 7:
            raise exc.DLMSException(F"Parameter got wrong index")

    @property
    def has_index(self) -> bool:
        return len(self.value) > 6

    @property
    @deprecated("use obis")
    def ln(self) -> bytes:
        """Logical Name"""
        return self.value[:6]

    def is_method(self) -> bool:
        return self.value[6] == 1

    @property
    def i(self) -> int:
        """attribute or method index"""
        return self.value[7]

    def set_i(self, index: int, is_method: bool = False) -> Self:
        val = Index.pack(is_method, index)
        if len(self.value) == 6:
            tmp = self.value + val
        else:
            tmp = bytearray(self.value)
            tmp[6:8] = val
            tmp = bytes(tmp)
        return self.__class__(tmp)

    def append_validate(self):
        if (l := len(self.value)) < 7:
            raise exc.DLMSException(F"Parameter must has index before")
        elif l % 2 != 0:
            raise exc.DLMSException(F"Can't append to Parameter with piece")

    def append(self, index: int) -> Self:
        """add new sequence(array or struct) index element"""
        self.append_validate()
        return self.__class__(self.value + pack(">H", index))

    def extend(self, *indexes: int):
        self.append_validate()
        return self.__class__(self.value + pack(F">{len(indexes)}H", *indexes))

    def pop(self) -> tuple[Optional[int], int, Self]:
        """
        :return piece, last index and parent Parameter
        ex.: Parameter("0.0.0.0.0.0:2 1/1/1 p3") => (1, Parameter("0.0.0.0.0.0:2 1/1"))
        """
        if self.has_piece():
            return self.value[-1], int.from_bytes(self.value[-3:-1]), self.__class__(self.value[:-3])
        else:
            return None, int.from_bytes(self.value[-2:]), self.__class__(self.value[:-2])

    def set_piece(self, index: int) -> Self:
        """add new sequence(array or struct) index element"""
        if len(self.value) >= 7:
            return self.__class__(self.value + pack("B", index))
        else:
            raise exc.DLMSException(F"Parameter must has index before")

    def has_piece(self) -> bool:
        if (
            (l := len(self.value)) >= 9
            and l % 2 != 0
        ):
            return True
        else:
            return False

    @property
    def piece(self) -> Optional[int]:
        if self.has_piece():
            return self.value[-1]

    def clear_piece(self) -> Self:
        if self.has_piece():
            return self.__class__(self.value[:-1])

    def elements(self, start: int = 0) -> iter:
        """return: index elements nested in attribute, started with"""
        for i in range(8 + start, 8 + 2 * self.n_elements, 2):
            res = int.from_bytes(self.value[i:i+2], "big")
            yield res

    @property
    def last_element(self) -> int:
        """:return last element index"""
        if self.n_elements == 0:
            raise ValueError("Parameter hasn't elements")
        if self.has_piece():
            val = self.value[-3: -1]
        else:
            val = self.value[-2:]
        return int.from_bytes(val, "big")

    @property
    def n_elements(self) -> int:
        """return: amount of elements nested in attribute"""
        return max(0, (len(self.value) - 8) // 2)

    def set(self,
            a: int = None,
            b: int = None,
            c: int = None,
            d: int = None,
            e: int = None,
            f: int = None
            ) -> Self:
        val = bytearray(self.value)
        if a is not None:
            val[0] = a
        if b is not None:
            val[1] = b
        if c is not None:
            val[2] = c
        if d is not None:
            val[3] = d
        if e is not None:
            val[4] = e
        if f is not None:
            val[5] = f
        return self.__class__(bytes(val))

    def __contains__(self, item: Self):
        return item.value in self.value

    def __getitem__(self, item) -> Optional[int]:
        if self.n_elements > 0:
            return unpack_from(">H", self.value, item*2 + 8)[0]
        else:
            return None

    @property
    def a(self) -> int:
        return self.value[0]

    @property
    def b(self) -> int:
        return self.value[1]

    @property
    def c(self) -> int:
        return self.value[2]

    @property
    def d(self) -> int:
        return self.value[3]

    @property
    def e(self) -> int:
        return self.value[4]

    @property
    def f(self) -> int:
        return self.value[5]

    @property
    def attr(self) -> Self:
        if self.has_index:
            return Parameter(self.value[:8])
        else:
            raise exc.DLMSException(F"Parameter must has index before")

    @property
    def obj(self) -> 'Parameter':
        return Parameter(self.value[:6])

    @property
    def obis(self) -> OBIS:
        return OBIS(self.value[:6])



from dataclasses import dataclass
from typing import TypeVar, Generic
from .cosem_interface_classes import Parameter
from .types import cdt

T = TypeVar("T")


@dataclass(frozen=True)
class ParValues(Generic[T]):
    par:  Parameter
    data: T

    def __getitem__(self, item):
        if item == 0:
            return self.par
        elif item == 1:
            return self.data
        else:
            raise StopIteration

    def __str__(self):
        return F"{self.par} - {self.data}"


@dataclass(frozen=True)
class ParData(ParValues):
    data: cdt.CommonDataType



"""DLMS UA 1000-1 Ed. 14"""
from . import ver0
from typing import Type, Iterator
from ... import cosem_interface_classes
from ...relation_to_OBIS import get_name
from ... import exceptions as exc
from ..__class_init__ import *
from ...types.implementations import integers, arrays, structs


class CaptureObjects(cdt.Array):
    """ Specifies the list of capture objects """
    TYPE = structs.CaptureObjectDefinition


class FromEntry(cdt.DoubleLongUnsigned, min=1):
    """ Access selector value for selective access to the object_list attribute """


class EntryDescriptor(cdt.Structure):
    """ Only buffer elements corresponding to the entry_descriptor shall be returned in the response.
    NOTE: from_entry and to_entry identify the lines, from_selected_value to_selected_value identify the columns of the buffer to be retrieved. """
    DEFAULT = (1, 0, 1, 0)
    from_entry: FromEntry
    to_entry: cdt.DoubleLongUnsigned
    from_selected_value: cdt.LongUnsigned
    to_selected_value: cdt.LongUnsigned


class AccessSelector(ut.Unsigned8):
    """ Unsigned8 1..2. Default is 2 for read all buffer """
    def __init__(self, value: int | str | ut.Unsigned8 = 2):
        super(AccessSelector, self).__init__(value)
        if int(self) not in (1, 2):
            raise ValueError(F'The {self.__class__.__name__} got {int(self)}, expected 1..2')


class ProfileGeneric(ver0.ProfileGeneric):
    """4.3.6 Profile generic"""
    VERSION = Version.V1
    A_ELEMENTS = (ic.ICAElement("buffer", arrays.SelectionAccess, classifier=ic.Classifier.DYNAMIC),
                  ic.ICAElement("capture_objects", CaptureObjects),
                  ver0.ProfileGeneric.get_attr_element(4),
                  ver0.ProfileGeneric.get_attr_element(5),
                  ic.ICAElement("sort_object", structs.CaptureObjectDefinition),
                  ic.ICAElement("entries_in_use", cdt.DoubleLongUnsigned, 0, default=0, classifier=ic.Classifier.DYNAMIC),
                  ic.ICAElement("profile_entries", cdt.DoubleLongUnsigned, 1, default=1))
    M_ELEMENTS = (
        ver0.ProfileGeneric.get_meth_element(1),
        ver0.ProfileGeneric.get_meth_element(2)
    )

    def characteristics_init(self):
        self.set_attr(ver0.BUFFER, None)

        # todo remove it
        self.buffer.register_cb_preset(lambda _: self.__create_buffer_struct_type())  # value not used for creating struct type

        self._cbs_attr_post_init.update({ver0.CAPTURE_OBJECTS: self.__create_buffer_struct_type,
                                         ver0.SORT_OBJECT: self.__create_selective_access_descriptor})

        self.buffer_capture_objects = self.capture_objects
        """ objects for buffer. Change with access_selection """

    @property
    def buffer(self) -> arrays.SelectionAccess:
        return self.get_attr(2)

    @property
    def capture_objects(self) -> CaptureObjects:
        return self.get_attr(3)

    @property
    def sort_object(self) -> structs.CaptureObjectDefinition:
        return self.get_attr(6)

    def get_attr_descriptor(self,
                            value: int,
                            with_selection: bool = False) -> ut.CosemAttributeDescriptor | ut.CosemAttributeDescriptorWithSelection:
        """ with selection for object_list. TODO: Copypast AssociationLN"""
        descriptor: ut.CosemAttributeDescriptor = super(ProfileGeneric, self).get_attr_descriptor(value)
        if value == ver0.BUFFER and with_selection:
            if self.attr_descriptor_with_selection is None:
                self.__create_selective_access_descriptor()
            return self.attr_descriptor_with_selection((descriptor.contents, self.buffer.selective_access.contents))
        else:
            return descriptor

    def __create_buffer_struct_type(self):
        """ TODO: more refactoring !!! """
        # rename CaptureObjectDefinition's and adding object if it absense in collection
        if self.buffer.selective_access is None:
            self.__create_selective_access_descriptor()
        if self.capture_objects is None:
            raise ValueError(F"{self}: create buffer Struct type, not set <capture_object> attribute. Need initiate <capture_objects> before")
        for el_value in self.capture_objects:
            el_value: structs.CaptureObjectDefinition
            obj = self.collection.add_if_missing(class_id=ut.CosemClassId(el_value.class_id.contents),
                                                 version=None,
                                                 logical_name=el_value.logical_name)
            el_value.set_name(self.collection.get_name_and_type(el_value)[0][-1])
        match self.buffer.selective_access:
            case ut.SelectiveAccessDescriptor() as desc:
                match int(desc.access_selector):
                    # case 0:                                                 self.buffer_capture_objects = self.capture_objects
                    case 1 if len(desc.access_parameters.selected_values) == 0: self.buffer_capture_objects = self.capture_objects
                    case 1:                                                     self.buffer_capture_objects = desc.access_parameters.selected_values
                    case 2:
                        from_selected_value = int(desc.access_parameters.from_selected_value)-1
                        to_selected_value = int(desc.access_parameters.to_selected_value)
                        if to_selected_value == 0:
                            to_selected_value = len(self.capture_objects)
                        self.buffer_capture_objects = self.capture_objects[from_selected_value:to_selected_value]
                    case _ as err:                                                raise ValueError(F'access_selection out of range, got {err}, must be (0..2)')
            case None:
                self.clear_attr(ver0.CAPTURE_OBJECTS)
                self._cbs_attr_post_init[ver0.CAPTURE_OBJECTS] = self.__create_buffer_struct_type
                raise exc.EmptyObj(F"need set <sort_object> before for {self}")
        buffer_elements: list[cdt.StructElement] = list()
        for el_value in self.buffer_capture_objects:
            names, type_ = self.collection.get_name_and_type(el_value)
            buffer_elements.append(cdt.StructElement(NAME=". ".join(names), TYPE=type_))

        class Entry(cdt.Structure):
            """4.3.6 Profile generic: entry"""
            ELEMENTS = tuple(buffer_elements)

        self.buffer.set_type(Entry)

    def __create_selective_access_descriptor(self):
        """ Available after got sort object. TODO: need rewrite. maybe replace to collection level. Wrong used sort_obj, it can be any element from capture_objects"""
        if self.sort_object is None:
            raise exc.EmptyObj(F"<sort object> is empty")
        sort_obj: ic.COSEMInterfaceClasses = self.collection.get_object(self.sort_object.logical_name)
        if sort_obj.CLASS_ID.contents == self.sort_object.class_id.contents:
            value_type: Type[cdt.CommonDataType] = sort_obj.get_attr_data_type(int(self.sort_object.attribute_index))
        else:
            exc.NoObject(F"got {self.sort_object.class_id=}, expected {sort_obj.CLASS_ID=} from collection")

        class RangeDescriptor(cdt.Structure):
            # cb_preset = TODO: make check 'selected_values' from self.capture_objects or
            # cb_post_set = TODO: make check 'selected_values' from self.capture_objects
            DEFAULT = b'\x02\x04\x02\x04\x12\x00\x01\x09\x06\x00\x00\x01\x00\x00\xff\x0f\x02\x12\x00\x00\x09\x0c\x07\xe4\x01\x01\xff\xff\xff\xff\xff\x80\x00\xff' \
                      b'\x09\x0c\x07\xe4\x01\x02\xff\xff\xff\xff\xff\x80\x00\xff\x01\x00'
            restricting_object: structs.CaptureObjectDefinition
            from_value: value_type
            to_value: value_type
            selected_values: CaptureObjects

        class Data(ut.Data):
            restricting_object: structs.CaptureObjectDefinition
            from_value: cdt.SimpleDataType
            to_value: cdt.SimpleDataType
            selected_values: CaptureObjects
            from_entry: FromEntry
            to_entry: cdt.DoubleLongUnsigned
            from_selected_value: cdt.LongUnsigned
            to_selected_value: cdt.LongUnsigned
            ELEMENTS = {1: ut.SequenceElement('range_descriptor', RangeDescriptor),
                        2: ut.SequenceElement('entry_descriptor', EntryDescriptor)}

        class SelectiveAccessDescriptor(ut.SelectiveAccessDescriptor):
            access_selector: AccessSelector
            access_parameters: Data
            ELEMENTS = (ut.SequenceElement('access_selector', AccessSelector),
                        ut.SequenceElement('access_parameters', Data))

        class CosemAttributeDescriptorWithSelection(ut.CosemAttributeDescriptorWithSelection):
            access_selection: SelectiveAccessDescriptor
            ELEMENTS = (ut.SequenceElement('cosem_attribute_descriptor', ut.CosemAttributeDescriptor),
                        ut.SequenceElement('access_selection', SelectiveAccessDescriptor))

        self.attr_descriptor_with_selection = CosemAttributeDescriptorWithSelection
        self.buffer.selective_access = SelectiveAccessDescriptor()

    def get_capture_object_names(self) -> list[str]:
        """ return all capture object names from collection """
        if self.capture_objects is None:
            raise ValueError(F'{self}: Empty capture objects')
        else:
            definition: structs.CaptureObjectDefinition
            ret = list()
            for definition in self.capture_objects:
                ret.append(get_name(definition.logical_name))
            return ret

    def get_buffer_objects(self) -> list[cosem_interface_classes.cosem_interface_class.COSEMInterfaceClasses]:
        """ get objects of current buffer container """
        return [self.collection.get(obj_def.logical_name.contents) for obj_def in self.buffer_capture_objects]

    # todo remove it
    def get_index_with_attributes(self, in_init_order: bool = False) -> Iterator[tuple[int, cdt.CommonDataType | None]]:
        """ override common method """
        return iter(((1, self.logical_name),
                    (6, self.sort_object),
                    (3, self.capture_objects),
                    (2, self.buffer),
                    (4, self.capture_period),
                    (5, self.sort_method),
                    (7, self.entries_in_use),
                    (8, self.profile_entries)))



from . import ver0
from ..__class_init__ import *
from ...types import choices


class SecurityPolicyVer1(cdt.IntegerFlag, cdt.Enum):
    """ security_policy"""


class CertificateEntity(cdt.Enum, elements=(0, 1, 2, 3)):
    """TODO:"""


class CertificateType(cdt.Enum, elements=(0, 1, 2, 3)):
    """TODO:"""


class SecuritySuite(ver0.SecuritySuite, elements=(0, 1, 2)):
    """Version 0 extension"""
    AES_GCM_128_AUT_ENCR_ECDSA_P_256_DIG_SIGN_ECDH_P_256_KEY_AGR_SHA_256_HASH_V44_COMPR_AND_AES_128_KEY_WRAP = 1
    AES_GCM_256_AUT_ENCR_ECDSA_P_384_DIG_SIGN_ECDH_P_384_KEY_AGR_SHA_384_HASH_V44_COMPR_AND_AES_256_KEY_WRAP = 2


class CertificateInfo(cdt.Structure):
    """ TODO: """
    certificate_entity: CertificateEntity
    certificate_type: CertificateType
    serial_number: cdt.OctetString
    issuer: cdt.OctetString
    subject: cdt.OctetString
    subject_alt_name: cdt.OctetString


class Certificates(cdt.Array):
    """Carries information on the X.509 v3 Certificates available and stored in the server"""
    TYPE = CertificateInfo


class KeyID(cdt.Enum, elements=(0, 1, 2, 3)):
    """Version 0 extension"""


class KeyTransferData(cdt.Structure):
    """ TODO: """
    key_id: KeyID
    key_wrapped: cdt.OctetString


class KeyTransfer(cdt.Array):
    """ Array of key_transfer_data """
    TYPE = KeyTransferData


class KeyAgreementData(cdt.Structure):
    """ TODO: """
    key_id: KeyID
    key_data: cdt.OctetString


class KeyAgreement(cdt.Array):
    """ Array of key_agreement_data """
    TYPE = KeyAgreementData


class KeyPair(cdt.Enum, elements=(0, 1, 2)):
    """TODO:"""


class CertificateIdentificationByEntity(cdt.Structure):
    """ TODO: """
    certificate_entity: CertificateEntity
    certificate_type: CertificateType
    system_title: cdt.OctetString


class CertificateIdentificationBySerial(cdt.Structure):
    """ TODO: """
    serial_number: cdt.OctetString
    issuer: cdt.OctetString


class CertificateIdentificationType(cdt.Enum, elements=(0, 1)):
    """TODO:"""


class CertificationIdentificationOption(ut.CHOICE):
    TYPE = cdt.Structure
    ELEMENTS = {0: ut.SequenceElement('by entity', CertificateIdentificationByEntity),
                1: ut.SequenceElement('by serial', CertificateIdentificationBySerial)}


certification_identification_option = CertificationIdentificationOption()


class CertificateIdentification(choices.StructureMixin, cdt.Structure):
    """Override several methods of cdt.Structure. It limited Structure."""
    certificate_identification_type: CertificateIdentificationType
    certification_identification_options: certification_identification_option


class SecuritySetup(ver0.SecuritySetup):
    VERSION = Version.V1
    A_ELEMENTS = (ic.ICAElement("security_policy", SecurityPolicyVer1),
                  ic.ICAElement("security_suite", SecuritySuite),
                  ver0.SecuritySetup.get_attr_element(4),
                  ver0.SecuritySetup.get_attr_element(5),
                  ic.ICAElement("certificates", Certificates, classifier=ic.Classifier.DYNAMIC))

    M_ELEMENTS = (ic.ICMElement("security_activate", SecurityPolicyVer1),
                  ic.ICMElement("key_transfer", KeyTransfer),
                  ic.ICMElement("key_agreement", KeyAgreement),
                  ic.ICMElement("generate_key_pair", KeyPair),
                  ic.ICMElement("generate_certificate_request", KeyPair),
                  ic.ICMElement("import_certificate", cdt.OctetString),
                  ic.ICMElement("export_certificate", CertificateIdentification),
                  ic.ICMElement("remove_certificate", CertificateIdentification))

    def characteristics_init(self):
        self.set_attr(2, None)
        self.set_attr(3, None)
        self.set_attr(6, None)

    @property
    def security_policy(self) -> SecurityPolicyVer1:
        return self.get_attr(2)

    @property
    def security_suite(self) -> SecuritySuite:
        return self.get_attr(3)

    @property
    def certificates(self) -> Certificates:
        return self.get_attr(6)

    @property
    def security_activate(self) -> SecurityPolicyVer1:
        return self.get_meth(1)

    @property
    def key_transfer(self) -> KeyTransferData:
        return self .get_meth(2)

    def key_agreement(self) -> KeyAgreement:
        return self .get_meth(3)

    def generate_certificate_request(self) -> KeyPair:
        return self .get_meth(4)

    def import_certificate(self) -> cdt.OctetString:
        return self .get_meth(5)

    def export_certificate(self) -> CertificateIdentification:
        return self .get_meth(6)

    def remove_certificate(self) -> CertificateIdentification:
        return self .get_meth(7)



from __future__ import annotations
from functools import lru_cache
from abc import ABC, abstractmethod
from typing import Type, Any, Callable
from dataclasses import dataclass
from ..types import common_data_types as cdt
from ..exceptions import DLMSException
from ..config_parser import get_values


class UserfulTypesException(DLMSException):
    """override DLMSException"""


class _String(ABC):
    LENGTH: int | None

    def __init__(self, value: bytes | bytearray | str | int | tuple | UsefulType = None):
        match value:
            case None:                                                        self.__dict__["contents"] = bytes(self.LENGTH)
            case bytes() if self.LENGTH is None or self.LENGTH <= len(value): self.__dict__["contents"] = value[:self.LENGTH]
            # case bytes() if self.LENGTH <= len(value):                       self.contents = value[:self.LENGTH]
            case bytes():            raise ValueError(F'Length of contents for {self.__class__.__name__} must be at least {self.LENGTH}, but got {len(value)}')
            case tuple():
                if len(value) == self.LENGTH:
                    self.__dict__["contents"] = bytes(value)
                else:
                    raise ValueError(F"in {self.__class__.__name__} with {value=} got length: {len(value)}, expect {self.LENGTH}")
            case bytearray():                                                 self.__dict__["contents"] = bytes(value)  # Attention!!! changed method content getting from bytearray
            case str():                                                       self.__dict__["contents"] = self.from_str(value)
            case int():                                                       self.__dict__["contents"] = self.from_int(value)
            case UsefulType():                                                self.__dict__["contents"] = value.contents  # TODO: make right type
            case _:                                                           raise ValueError(F'Error create {self.__class__.__name__} with value {value}')

    @abstractmethod
    def __len__(self):
        """ define in subclasses """


class OCTET_STRING(_String):
    """ An ordered sequence of octets (8 bit bytes) """

    def from_str(self, value: str) -> bytes:
        """ input as hex code """
        return bytes.fromhex(value)

    def from_int(self, value: int) -> bytes:
        """ Convert with recursion. Maximum convert length is 32 """
        def to_bytes_with(length_):
            try:
                return int.to_bytes(value, length_, 'big')
            except OverflowError:
                if length_ > 31:
                    raise ValueError(F'Value {value} is big to convert to bytes')
                return to_bytes_with(length_+1)
        length = 1
        return to_bytes_with(length)

    def __str__(self):
        return self.contents.hex(' ')

    def __len__(self):
        return len(self.contents)

    def __getitem__(self, item):
        return self.contents[item]

    # TODO: maybe remove as redundante?
    def decode(self) -> bytes:
        """ decode to build in bytes type """
        return self.contents

    # TODO: maybe remove as redundante?
    def to_str(self, encoding: str = 'cp1251') -> str:
        """ decode to cp1251 by default, replace to '?' if unsupported """
        temp = list()
        for i in self.contents:
            temp.append(i if i > 32 else 63)
        return bytes(temp).decode(encoding)


class CHOICE(ABC):
    """ TODO: with cdt.CHOICE """
    ELEMENTS: dict[int, SequenceElement | dict[int, SequenceElement]]

    @property
    @abstractmethod
    def TYPE(self) -> Any:
        """ return valid types """

    def __init_subclass__(cls, **kwargs):
        if hasattr(cls, 'ELEMENTS'):
            for el in cls.ELEMENTS.values():
                if isinstance(el, dict):
                    """pass, maybe it is for cst.AnyTime"""
                elif issubclass(el.TYPE, cls.TYPE):
                    """ type in order """
                else:
                    raise ValueError(F'For {cls.__name__} got type {el.TYPE.__name__} with {el.NAME=}, expected {cls.TYPE.__name__}')
        else:
            """ subclass with type carry initiate """

    def __getitem__(self, item: int) -> SequenceElement:
        return self.ELEMENTS[item]

    @property
    def NAME(self) -> str:
        return F'CHOICE[{len(self)}]'

    def __len__(self):
        return len(self.ELEMENTS)

    def is_key(self, value: int) -> bool:
        return value in self.ELEMENTS.keys()

    @classmethod
    def parse(cls, value: cdt.Transcript) -> cdt.CommonDataType:
        """get instance by pattern: <tag>:<value>"""
        tag, value = value.split(":", maxsplit=1)
        if (d_t := cls.ELEMENTS.get(int(tag))) is None:
            raise UserfulTypesException(F"for {cls.__name__} got {cdt.CommonDataType.__name__}: {cdt.TAG(int(tag).to_bytes(1))}; "
                                        F"expected: {', '.join(map(lambda el: el.NAME, cls.ELEMENTS.values()))}")
        else:
            return d_t.TYPE.parse(value)

    def __call__(self,
                 value: bytes | int = None,
                 force: bool = False) -> cdt.CommonDataType:
        """ get instance from encoding or tag(with default value). For CommonDataType only """
        try:
            match value:
                case bytes() as encoding:
                    match self.ELEMENTS[encoding[0]]:
                        case SequenceElement() as el: return el.TYPE(encoding)
                        case dict() as ch:
                            if encoding[1] in ch.keys():
                                return ch[encoding[1]].TYPE(encoding)  # use for choice cst.Time | DateTime | Date as OctetString
                            else:
                                raise ValueError(F"got type with tag: {encoding[0]} and length: {encoding[1]}, expected length {tuple(ch.keys())}")
                        case err:                     raise ValueError(F"got {err.__name__}, expected {SequenceElement.__name__} or {dict.__name__}")
                case int() if force:                  return cdt.get_common_data_type_from(value.to_bytes(1, "big"))()
                case int() as tag:                    return self.ELEMENTS[tag].TYPE()
                case None:                            return tuple(self.ELEMENTS.values())[0].TYPE()
                case error:                           raise ValueError(F'Unknown value type {error}')
        except KeyError as e:
            raise UserfulTypesException(F"for {self.__class__.__name__} got {cdt.CommonDataType.__name__}: {cdt.TAG(e.args[0].to_bytes(1))}; expected: {', '.join(map(lambda el: el.NAME, self.ELEMENTS.values()))}")

    def __get_elements(self) -> list[SequenceElement]:
        """all elements with nested values"""
        elements = list()
        for el in self.ELEMENTS.values():
            match el:
                case SequenceElement():
                    elements.append(el)
                case dict() as dict_el:
                    elements.extend(dict_el.values())
                case err:
                    raise ValueError(F"unknown CHOICE element type {err}")
        return elements

    def get_types(self) -> tuple[Type[cdt.CommonDataType]]:
        """ Use in setter attribute.value for validate """
        return tuple((seq_el.TYPE for seq_el in self.__get_elements()))

    def __str__(self):
        return F'{CHOICE}: {", ".join((el.NAME for el in self.__get_elements()))}'


def get_instance_and_context(meta: Type[UsefulType], value: bytes) -> tuple[UsefulType, bytes]:
    instance = meta(value)
    return instance, value[len(instance.contents):]


@dataclass(frozen=True)
class SequenceElement:
    NAME: str
    TYPE: Type[UsefulType | SEQUENCE | CHOICE | cdt.CommonDataType]

    def __str__(self):
        return F'{self.NAME}: {self.TYPE.__name__}'


class SEQUENCE(ABC):
    """ TODO: """
    ELEMENTS: tuple[SequenceElement | SEQUENCE, ...]
    values: list[UsefulType]

    def __init__(self, value: bytes | tuple | list | None | SEQUENCE = None):
        self.__dict__['values'] = [None] * len(self.ELEMENTS)
        match value:
            case tuple() | list():
                if len(value) != len(self):
                    raise ValueError(F'Struct {self.__class__.__name__} got length:{len(value)}, expected length:{len(self)}')
                self.from_tuple(value)
            case bytes():           self.from_bytes(value)
            case None:              self.from_default()
            case self.__class__():  self.from_bytes(value.contents)
            case _:                 raise TypeError(F'Value: "{value}" not supported')

    def from_default(self):
        for i in range(len(self)):
            self.values[i] = self.ELEMENTS[i].TYPE()

    def from_bytes(self, value: bytes | bytearray):
        for i, element in enumerate(self.ELEMENTS):
            self.values[i], value = get_instance_and_context(element.TYPE, value)

    def from_tuple(self, value: tuple | list):
        for i, val in enumerate(value):
            self.values[i] = self.ELEMENTS[i].TYPE(val)

    def __str__(self):
        return F'{{{", ".join(map(lambda val: F"{val[0].NAME}: {val[1]}", zip(self.ELEMENTS, self.values)))}}}'

    def __repr__(self):
        return F'{self.__class__.__name__}(({(", ".join(map(str, self.values)))}))'

    def __get_index(self, name: str) -> int | None:
        """ get index by name. Return None if not found """
        for i, element in enumerate(self.ELEMENTS):
            if element.NAME == name:
                return i
        else:
            return None

    def __setattr__(self, key, value: UsefulType):
        match self.__get_index(key):
            case int() as i if isinstance(value, self.ELEMENTS[i]):                self.values[i] = value
            case int() as i:                raise ValueError(F'Try assign {key} Type got {value.__class__.__name__}, expected {self.ELEMENTS[i].NAME}')
            case _:                raise ValueError(F'Unsupported change: {key}')

    def __getattr__(self, item: str) -> UsefulType:
        match self.__get_index(item):
            case int() as i: return self.values[i]
            case _:          return self.__dict__[item]

    def __getitem__(self, item: str | int) -> UsefulType:
        """ get element by index or name """
        match item:
            case str() as name:  return self.values[self.__get_index(name)]
            case int() as index: return self.values[index]
            case _:              raise ValueError(F'Unsupported type {item.__class__}')

    def __setitem__(self, key: int, value: UsefulType):
        """ set data to element by index """
        if isinstance(value, self.ELEMENTS[key].TYPE):
            self.values[key] = value
        else:
            raise ValueError(F'Type got {value.__class__.__name__}, expected {self.ELEMENTS[key].TYPE}')

    def __len__(self):
        return len(self.ELEMENTS)

    @property
    def contents(self) -> bytes:
        return b''.join(el.contents for el in self.values)

    @property
    def NAME(self) -> str:
        return F'{self.__class__.__name__}[{len(self)}]'


class UsefulType(ABC):
    """"""
    contents: bytes
    __match_args__ = ('contents',)
    cb_post_set: Callable
    cb_preset: Callable

    @abstractmethod
    def __init__(self, value):
        """ constructor """

    def __eq__(self, other: UsefulType):
        match other:
            case self.__class__(self.contents): return True
            case _:                             return False

    def set_contents_from(self, value: UsefulType | bytes | bytearray | str | int | bool | None):
        new_value = self.__class__(value)
        if hasattr(self, 'cb_preset'):
            self.cb_preset(new_value)
        self.__dict__['contents'] = new_value.contents
        if hasattr(self, 'cb_post_set'):
            self.cb_post_set()

    def __setattr__(self, key, value):
        match key:
            case 'contents' as prop if hasattr(self, 'contents'): raise ValueError(F"Don't support set {prop}")
            case _: super().__setattr__(key, value)


class DigitalMixin(ABC):
    """ Default value is 0 """
    SIGNED: bool
    LENGTH: int
    contents: bytes

    def __init__(self, value: bytes | bytearray | str | int | DigitalMixin = None):
        match value:
            case bytes() if self.LENGTH <= len(value): self.__dict__["contents"] = value[:self.LENGTH]
            case bytes():            raise ValueError(F'Length of contents for {self.__class__.__name__} must be at least {self.LENGTH}, but got {len(value)}')
            case bytearray():                          self.__dict__["contents"] = bytes(value)  # Attention!!! changed method content getting from bytearray
            case str('-') if self.SIGNED:              self.__dict__["contents"] = bytes(self.LENGTH)
            case int():                                self.__dict__["contents"] = self.from_int(value)
            case str():                                self.__dict__["contents"] = self.from_str(value)
            case None:                                 self.__dict__["contents"] = bytes(self.LENGTH)
            case self.__class__():                     self.__dict__["contents"] = value.contents
            case _:                                    raise ValueError(F'Error create {self.__class__.__name__} with value: {value}')

    def from_int(self, value: int | float) -> bytes:
        try:
            return int(value).to_bytes(self.LENGTH, 'big', signed=self.SIGNED)
        except OverflowError:
            raise ValueError(F'value {value} out of range')

    def from_str(self, value: str) -> bytes:
        return self.from_int(float(value))

    def __int__(self):
        """ return the build in integer type """
        return int.from_bytes(self.contents, 'big', signed=self.SIGNED)

    def __str__(self):
        return str(int(self))

    def __repr__(self):
        return F'{self.__class__.__name__}({self})'

    def __gt__(self, other: DigitalMixin):
        match other:
            case DigitalMixin(): return int(self) > int(other)
            case _:          raise TypeError(F'Compare type is {other.__class__}, expected Digital')

    def __len__(self) -> int:
        return self.LENGTH

    def __hash__(self):
        return int(self)

    def validate_from(self, value: str, cursor_position: int) -> tuple[str, int]:
        """ return validated value and cursor position. Use in Entry. TODO: remove it, make better """
        type(self)(value=value)
        return value, cursor_position


class Integer8(DigitalMixin, UsefulType):
    """ INTEGER(-127…128) """
    LENGTH = 1
    SIGNED = True


class Integer16(DigitalMixin, UsefulType):
    """ INTEGER(-32 768...32 767) """
    LENGTH = 2
    SIGNED = True


class Integer32(DigitalMixin, UsefulType):
    """ INTEGER(-2 147 483 648...2 147 483 647) """
    LENGTH = 4
    SIGNED = True


class Integer64(DigitalMixin, UsefulType):
    """ INTEGER(-2^63...2^63-1) """
    LENGTH = 8
    SIGNED = True


class Unsigned8(DigitalMixin, UsefulType):
    """ INTEGER(0...255) """
    LENGTH = 1
    SIGNED = False
    # __match_args__ = ('contents',)


class Unsigned16(DigitalMixin, UsefulType):
    """ INTEGER(0...65 535) """
    LENGTH = 2
    SIGNED = False


class Unsigned32(DigitalMixin, UsefulType):
    """ INTEGER(0...4 294 967 295) """
    LENGTH = 4
    SIGNED = False


class Unsigned64(DigitalMixin, UsefulType):
    """ INTEGER(0...264-1) """
    LENGTH = 8
    SIGNED = False


class CosemClassId(Unsigned16):
    """ Identification code of the IC (range 0 to 65 535). The class_id of each object is retrieved together with the logical name by reading the object_list attribute of an
    “Association LN” / ”Association SN” object.
        - class_id-s from 0 to 8 191 are reserved to be specified by the DLMS UA.
        - class_id-s from 8 192 to 32 767 are reserved for manufacturer specific ICs.
        - class_id-s from 32 768 to 65 535 are reserved for user group specific ICs.
    The DLMS UA reserves the right to assign ranges to individual manufacturers or user groups. """

    def __str__(self):
        if _class_names:
            return _class_names.get(self, repr(self))
        else:
            return repr(self)

    def __repr__(self):
        return F"{self.__class__.__name__}({int(self)})"


_class_names = {CosemClassId(k): v for k, v in class_names.items()} if (class_names := get_values("DLMS", "class_name")) else None


class CosemObjectInstanceId(OCTET_STRING, UsefulType):
    LENGTH = 6

    def __str__(self):
        return F"\"{'.'.join(map(str, self.contents))}\""

    def from_str(self, value: str) -> bytes:
        """ create logical_name: octet_string from string type ddd.ddd.ddd.ddd.ddd.ddd, ex.: 0.0.1.0.0.255 """
        raw_value = bytes()
        for typecast, separator in zip((self.__from_group_A, )*5+(self.__from_group_F, ), ('.', '.', '.', '.', '.', ' ')):
            try:
                element, value = value.split(separator, 1)
            except ValueError:
                element, value = value, ''
            raw_value += typecast(element)
        return raw_value

    def __from_group_A(self, value: str) -> bytes:
        if isinstance(value, str):
            if value == '':
                return b'\x00'
            try:
                return int(value).to_bytes(1, 'big')
            except OverflowError:
                raise ValueError(F'Int too big to convert {value}')
        else:
            raise TypeError(F'Unsupported type validation from string, got {value.__class__}')

    def __from_group_F(self, value: str) -> bytes:
        if isinstance(value, str):
            if value == '':
                return b'\xff'
            try:
                return int(value).to_bytes(1, 'big')
            except OverflowError:
                raise ValueError(F'Int too big to convert {value}')
        else:
            raise TypeError(F'Unsupported type validation from string, got {value.__class__}')


class CosemObjectAttributeId(Integer8):
    """ TODO """

    @lru_cache(14)  # for test
    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)


class CosemObjectMethodId(Integer8):
    """ TODO """


class AccessSelectionParameters(Unsigned8):
    """ Unsigned8(0..1) """

    def __init__(self, value: int | str | Unsigned8 = 1):
        super(AccessSelectionParameters, self).__init__(value)
        if int(self) > 1 or int(self) < 0:
            raise ValueError(F'The {self.__class__.__name__} got {self}, expected 0..1')


class CosemAttributeDescriptor(SEQUENCE):
    class_id: CosemClassId
    instance_id: CosemObjectInstanceId
    attribute_id: CosemObjectAttributeId
    access_selection_parameters: AccessSelectionParameters
    ELEMENTS = (SequenceElement('class_id', CosemClassId),
                SequenceElement('instance_id', CosemObjectInstanceId),
                SequenceElement('attribute_id', CosemObjectAttributeId))

    def __init__(self, value: bytes | tuple | list = None):
        super(CosemAttributeDescriptor, self).__init__(value)
        self.__dict__['access_selection_parameters'] = AccessSelectionParameters(0)

    @property
    def contents(self) -> bytes:
        """ Always contain Access_Selection_Parameters. DLMS UA 1000-2 Ed.9 Excerpt 9.3.9.1.3 """
        return super(CosemAttributeDescriptor, self).contents + self.access_selection_parameters.contents


class CosemMethodDescriptor(SEQUENCE):
    class_id: CosemClassId
    instance_id: CosemObjectInstanceId
    method_id: CosemObjectMethodId
    ELEMENTS = (SequenceElement('class_id', CosemClassId),
                SequenceElement('instance_id', CosemObjectInstanceId),
                SequenceElement('method_id', CosemObjectMethodId))

    def __init__(self, value: tuple[CosemClassId, CosemObjectInstanceId, CosemObjectMethodId]):
        super(CosemMethodDescriptor, self).__init__(value)


class Data(CHOICE, ABC):
    TYPE = cdt.CommonDataType


class SelectiveAccessDescriptor(SEQUENCE, ABC):
    """ Selective access specification always starts with an access selector, followed by an access-specific access parameter list.
    Specified IS/IEC 62056-53 : 2006, 7.4.1.6 Selective access """
    access_selector: Unsigned8
    access_parameters: cdt.CommonDataType
    ELEMENTS: tuple[SequenceElement, SequenceElement]

    def __init__(self, value: tuple | bytes | None = None):
        super(SelectiveAccessDescriptor, self).__init__(value)
        self.access_selector.cb_post_set = self.__validate_selector

    @property
    @abstractmethod
    def ELEMENTS(self) -> tuple[SequenceElement, SequenceElement]:
        """ return elements """

    def from_default(self):
        self.values[0] = self.ELEMENTS[0].TYPE()
        self.values[1] = self.ELEMENTS[1].TYPE.ELEMENTS[int(self.access_selector)].TYPE()

    def from_bytes(self, value: bytes | bytearray):
        self.values[0], value = get_instance_and_context(self.ELEMENTS[0].TYPE, value)
        self.values[1] = self.ELEMENTS[1].TYPE.ELEMENTS[int(self.access_selector)].TYPE(value)

    def from_tuple(self, value: tuple | list):
        self.values[0] = self.ELEMENTS[0].TYPE(value[0])
        self.values[1] = self.ELEMENTS[1].TYPE.ELEMENTS[int(self.access_selector)].TYPE(value[1])

    @property
    def contents(self) -> bytes:
        return self.access_selector.contents+self.access_parameters.encoding

    def __validate_selector(self):
        self.values[1] = self.ELEMENTS[1].TYPE.ELEMENTS[int(self.access_selector)].TYPE()
        print('change sel')

    def set_selector(self, index: int, value=None):
        """ set selector value from int type """
        self.access_selector.set_contents_from(index)
        if value:
            self.access_parameters.set(value)


class CosemAttributeDescriptorWithSelection(SEQUENCE):
    cosem_attribute_descriptor: CosemAttributeDescriptor
    access_selection: SelectiveAccessDescriptor
    ELEMENTS: tuple[CosemAttributeDescriptor, SequenceElement]

    def __init__(self, value: bytes | tuple | list = None):
        super(CosemAttributeDescriptorWithSelection, self).__init__(value)
        self.cosem_attribute_descriptor.access_selection_parameters.set_contents_from(1)

    @property
    @abstractmethod
    def ELEMENTS(self) -> tuple[SequenceElement, SequenceElement]:
        """ return elements. Need initiate in subclass """


class InvokeIdAndPriority(Unsigned8):

    def __init__(self, value: bytes | bytearray | str | int | DigitalMixin = 0b1100_0000):
        super(InvokeIdAndPriority, self).__init__(value)
        if self.contents[0] & 0b00110000:
            raise ValueError(F'For {self.__class__.__name__} set reserved bits')

    @classmethod
    def from_parameters(cls, invoke_id: int = 0,
                        service_class: int = 0,
                        priority: int = 0):
        instance = cls()
        instance.invoke_id = invoke_id
        instance.service_class = service_class
        instance.priority = priority
        return instance

    @property
    def invoke_id(self) -> int:
        return self.contents[0] & 0b0000_1111

    @invoke_id.setter
    def invoke_id(self, value: int):
        if value & 0b00001111:
            self.__dict__['contents'] = int((self.contents[0] & 0b1111_0000) | value).to_bytes(1, 'big')
        else:
            raise ValueError(F'Got {value} for invoke-id, expected 0..15')

    @property
    def service_class(self) -> int:
        """ 0: Unconfirmed or 1: Confirmed bit return """
        return (self.contents[0] >> 6) & 0b1

    @service_class.setter
    def service_class(self, value: int):
        match value:
            case 0 | 1: self.__dict__['contents'] = int((self.contents[0] & 0b1011_1111) | (value << 6)).to_bytes(1, 'big')
            case _:     raise ValueError(F'Got {value} for service_class, expected 0..1')

    @property
    def priority(self) -> int:
        """ 0: Normal or 1: High bit return """
        return (self.contents[0] >> 7) & 0b1

    @priority.setter
    def priority(self, value: int):
        match value:
            case 0 | 1: self.__dict__['contents'] = int((self.contents[0] & 0b0111_1111) | (value << 7)).to_bytes(1, 'big')
            case _:     raise ValueError(F'Got {value} for service_class, expected 0..1')

    def __str__(self):
        return F'priority: {"High" if self.contents[0] & 0b1000_0000 else "Normal"}, ' \
               F'service-class: {"Confirmed" if self.contents[0] & 0b0100_0000 else "Unconfirmed"}, ' \
               F'invoke-id: {self.invoke_id},'


if __name__ == '__main__':
    a = CosemObjectAttributeId(1)
    b = CosemObjectAttributeId(2)
    print(a > b)

    a = CosemClassId(cdt.LongUnsigned(1).contents)
    a = InvokeIdAndPriority()
    c = a.invoke_id
    d = a.service_class
    f = a.priority
    a.service_class = 1
    a.invoke_id = 14
    e = InvokeIdAndPriority.from_parameters(1, 1, 1)
    class AccessSelector(Unsigned8):
        """ Unsigned8 1..4 """
        def __init__(self, value: int | str | Unsigned8 = 1):
            super(AccessSelector, self).__init__(value)
            if int(self) > 4 or int(self) < 1:
                raise ValueError(F'The {self.__class__.__name__} got {self}, expected 1..4')


    class MyData(Data):
        ELEMENTS = {1: SequenceElement('0 a', cdt.NullData),
                    2: SequenceElement('1 d', cdt.Integer),
                    3: SequenceElement('second', cdt.ScalUnitType),
                    4: SequenceElement('3', cdt.Integer)}

    class My(SelectiveAccessDescriptor):
        access_selector: AccessSelector
        access_parameters: MyData
        ELEMENTS = (SequenceElement('access_selector', AccessSelector),
                    SequenceElement('access_parameters', MyData))

    ba = My()
    b = My((3, (10, 10)))
    b2 = b.access_selector
    b3 = b.access_parameters
    b4 = b.access_parameters.unit
    b_repr = My(b'\x03\x0f"')
    b.set_selector(3, 34)
    a = CosemAttributeDescriptor((1, '1.1.1.1.1.1', 1))
    a_repr = CosemAttributeDescriptor(b'\x00\x01\x01\x01\x01\x01\x01\x01\x01\x00')

    class MyWith(CosemAttributeDescriptorWithSelection):
        access_selection: My
        ELEMENTS = (SequenceElement('cosem_attribute_descriptor', CosemAttributeDescriptor),
                    SequenceElement('access_selection', My))

    c = MyWith()
    c_from = MyWith((a, b))
    c_repr = MyWith(b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00')
    print(a)
    c = CosemAttributeDescriptor(b'\x00\x01\x01\x01\x01\x01\x01\x01\x01\x00')
    print(c)



import unittest
from src.DLMS_SPODES.types import cdt, cst, ut
from src.DLMS_SPODES.cosem_interface_classes import collection, overview
from src.DLMS_SPODES.cosem_interface_classes import implementations as impl
# from .init_collection import col

class TestType(unittest.TestCase):

    def test_Data(self):
        obj = collection.Data("0.0.0.1.0.255")
        obj.set_attr(2, cdt.Unsigned(8).encoding)
        a = obj.get_attr(2)
        a.set(3)
        print(obj.value)

    def test_eventsData(self):
        col = collection.Collection()
        col.add(class_id=ut.CosemClassId(7), version=cdt.Unsigned(1), logical_name=cst.LogicalName.from_obis("0.0.99.98.4.255"))
        print(col)

    def test_ExternalEventData(self):
        col = collection.Collection(
            id_=collection.ID(
                man=b'KPZ',
                f_id=collection.ParameterValue(
                    par=b'',
                    value=cdt.OctetString("4d324d5f33").encoding),
                f_ver=collection.ParameterValue(
                    par=b'',
                    value=cdt.OctetString(bytearray(b'1.3.0')).encoding)
            )
        )
        col.spec_map = col.get_spec()
        print(col)
        col.add(class_id=ut.CosemClassId(1), version=cdt.Unsigned(0), logical_name=cst.LogicalName.from_obis("0.0.96.11.4.255"))
        obj = col.get_object("0.0.96.11.4.255")
        obj.set_attr(2, cdt.LongUnsigned(2).encoding)
        # obj.set_attr(2, 2)
        print(obj.value, obj.value.get_report())
        self.assertEqual(obj.value.get_report().msg, "(2) Магнитное поле - окончание", "report match")
        print(col)

    def test_DeviceIdObjects(self):
        obj = impl.data.DLMSDeviceIDObject("0.0.96.1.4.255")
        datatime = cdt.DateTime("01.01.20 11:00")
        obj.set_attr(2, datatime.encoding)

    def test_SealStatus(self):
        obj = impl.data.SealStatus("0.0.96.51.5.255")
        obj.set_attr(2, 0)
        for i in range(255):
            obj.value.set(i)
            print(i, obj.value.get_report())

    def test_Spodes_Version(self):
        obj = impl.data.SPODES3SPODESVersion("00 00 60 01 06 FF")
        obj.set_attr(2, "33 2e 30")
        rep = obj.value.get_report()
        print(rep)


import unittest
from src.DLMS_SPODES.types import cdt, cst, ut
from src.DLMS_SPODES.cosem_interface_classes import collection, overview


class TestType(unittest.TestCase):

    def test_ProfileGeneric(self):
        col = collection.Collection()
        col.add(col.add(class_id=ut.CosemClassId(15), version=cdt.Unsigned(1), logical_name=cst.LogicalName.from_obis('3.0.40.0.0.255')))
        col.add(class_id=ut.CosemClassId(8), version=cdt.Unsigned(0), logical_name=cst.LogicalName.from_obis('0.0.1.0.0.255'))
        col.add(class_id=ut.CosemClassId(47), version=cdt.Unsigned(0), logical_name=cst.LogicalName.from_obis("0.0.25.6.0.255"))
        new = col.add(class_id=ut.CosemClassId(1), version=cdt.Unsigned(0), logical_name=cst.LogicalName.from_obis('0.0.96.11.0.255'))
        new.set_attr(2, b'\x09\x02hi')
        reg = col.add(class_id=ut.CosemClassId(3), version=cdt.Unsigned(0), logical_name=cst.LogicalName.from_obis('1.0.12.7.0.255'))
        reg.set_attr(2, 6)
        col.add(class_id=ut.CosemClassId(3),
                   version=cdt.Unsigned(0),
                   logical_name=cst.LogicalName.from_obis('1.0.12.7.4.255')
                   ).set_attr(2, 6)
        col.add(class_id=ut.CosemClassId(1),
                   version=cdt.Unsigned(0),
                   logical_name=cst.LogicalName.from_obis('0.0.96.8.10.255')
                   ).set_attr(2, 6)
        col.add(class_id=ut.CosemClassId(1),
                   version=cdt.Unsigned(0),
                   logical_name=cst.LogicalName.from_obis('0.0.96.8.0.255')
                   ).set_attr(2, 6)
        inst = col.add(class_id=ut.CosemClassId(7), version=cdt.Unsigned(1), logical_name=cst.LogicalName.from_obis('0.0.99.13.0.255'))
        inst.set_attr(6, bytes.fromhex('020412000809060000010000ff0f02120000'))
        inst.set_attr(3, bytes.fromhex('01 02 02 04 12 00 2f 09 06 00 00 19 06 00 ff 0f 02 12 00 00 02 04 12 00 2f 09 06 00 00 19 06 00 ff 0f 06 12 00 03'))
        inst.set_attr(2, bytes.fromhex('01 03 02 02 0a 07 4d 65 67 61 46 6f 6e 11 1b 02 02 0a 0c 42 65 65 20 4c 69 6e 65 20 47 53 4d 11 3f 02 02 0a 03 4d 54 53 11 33'))
        # inst.set_attr(3, bytes.fromhex('0106020412000809060000010000ff0f02120000020412000109060000600b00ff0f021200000204120003090601000c0700ff0f021200000204120003090601000c0704ff0f0212000002041200010906000060080aff0f02120000020412000309060000600800ff0f02120000'))
        print(inst.get_capture_object_names())
        print(inst)

    def test_DisplayReadout(self):
        col = collection.Collection()
        col.set_manufacturer(b"KPZ")
        col.spec_map = col.get_spec()
        readout = col.add(class_id=overview.ClassID.PROFILE_GENERIC, version=overview.Version.V1, logical_name=cst.LogicalName.from_obis('0.0.21.0.1.255'))
        col.add(class_id=overview.ClassID.REGISTER, version=overview.Version.V0, logical_name=cst.LogicalName.from_obis('1.0.1.8.0.255'))
        col.add(class_id=overview.ClassID.REGISTER, version=overview.Version.V0, logical_name=cst.LogicalName.from_obis('1.0.1.8.1.255'))
        col.add(class_id=overview.ClassID.REGISTER, version=overview.Version.V0, logical_name=cst.LogicalName.from_obis('1.0.1.8.2.255'))
        col.add(class_id=overview.ClassID.REGISTER, version=overview.Version.V0, logical_name=cst.LogicalName.from_obis('1.0.1.8.3.255'))
        readout.set_attr(5, 1)
        readout.set_attr(3, bytes.fromhex("""01 04
        02 04 12 00 03 09 06 01 00 01 08 00 ff 0f 02 12 00 00 
        02 04 12 00 03 09 06 01 00 01 08 01 ff 0f 02 12 00 00 
        02 04 12 00 03 09 06 01 00 01 08 02 ff 0f 02 12 00 00 
        02 04 12 00 03 09 06 01 00 01 08 03 ff 0f 02 12 00 00"""))
        readout.set_attr(2, bytes.fromhex("01 01 02 51 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 00 03 00 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 01 03 00"))
        desc = readout.get_attr_descriptor(2)
        print(readout)

    def test_ver0(self):
        col = collection.Collection()
        col.spec_map = "SPODES_3"
        pg1 = col.add(
            class_id=overview.ClassID.PROFILE_GENERIC,
            version=overview.Version.V1,
            logical_name=cst.LogicalName("01 00 5e 07 04 ff")
        )
        pg2 = col.add(
            class_id=overview.ClassID.PROFILE_GENERIC,
            version=overview.Version.V1,
            logical_name=cst.LogicalName.from_obis("1.0.99.1.0.255")
        )
        pg3 = col.add(
            class_id=overview.ClassID.PROFILE_GENERIC,
            version=overview.Version.V1,
            logical_name=cst.LogicalName.from_obis("1.0.99.2.0.255")
        )
        pg4 = col.add_if_missing(
            class_id=overview.ClassID.PROFILE_GENERIC,
            version=overview.Version.V1,
            logical_name=cst.LogicalName("00 00 15 00 01 ff")
        )
        ver = col.find_version(overview.ClassID.PROFILE_GENERIC)
        print(col)
        col2 = col.copy()



import datetime
from .__class_init__ import *
from ..types.implementations import integers


class ClockStatus(cdt.Unsigned):
    """ interpreted as 8 bit string """

    # TODO: finish write as bit_string
    def __str__(self):
        value = int.from_bytes(self.contents, 'big')
        ret = ''
        if bool(value & 0b1):
            ret += ' invalid value'
        if bool(value & 0b10):
            ret += ' doubtful value'
        if bool(value & 0b100):
            ret += ' different clock base'
        if bool(value & 0b1000):
            ret += ' invalid clock status'
        if bool(value & 0b1000000):
            ret += ' daylight saving active'
        if ret == '':
            ret = 'empty'
        return ret


class DaylightSavingsDeviation(cdt.Integer, min=-120, max=120):
    """Contains the number of minutes by which the deviation in generalized time must be corrected at daylight savings begin.
    Deviation range of up to ± 120 min"""


class ClockBase(cdt.Enum, elements=(0, 1, 2, 3, 4, 5)):
    """ Defines where the basic timing information comes from. """


class PresetAdjustingTime(cdt.Structure):
    """ Presets the time to a new value (preset_time) and defines a validity_interval within which the new time can be activated """
    DEFAULT = b'\x02\x03\x19\x07\xe4\x01\x01\xff\xff\xff\xff\xff\x80\x00\xff' \
              b'\x19\x07\xe4\x01\x01\xff\xff\xff\xff\xff\x80\x00\xff\x19\x07\xe4\x01\x01\xff\xff\xff\xff\xff\x80\x00\xff'
    preset_time: cdt.DateTime
    validity_interval_start: cdt.DateTime
    validity_interval_end: cdt.DateTime


class ShiftTime(cdt.Long, min=-900, max=900):
    """ Limited Long -900..900 """


class TimeZone(cdt.Long):
    """"""


class Clock(ic.COSEMInterfaceClasses):
    """4.5.1 Clock"""
    CLASS_ID = ClassID.CLOCK
    VERSION = Version.V0
    A_ELEMENTS = (ic.ICAElement("time", cst.OctetStringDateTime, classifier=ic.Classifier.DYNAMIC),
                  ic.ICAElement("time_zone", TimeZone, -720, 840),
                  ic.ICAElement("status", ClockStatus, classifier=ic.Classifier.DYNAMIC),
                  ic.ICAElement("daylight_savings_begin", cst.OctetStringDateTime),
                  ic.ICAElement("daylight_savings_end", cst.OctetStringDateTime),
                  ic.ICAElement("daylight_savings_deviation", DaylightSavingsDeviation, -120, 120),
                  ic.ICAElement("daylight_savings_enabled", cdt.Boolean),
                  ic.ICAElement("clock_base", ClockBase))
    M_ELEMENTS = (ic.ICMElement("adjust_to_quarter", integers.Only0),
                  ic.ICMElement("adjust_to_measuring_period", integers.Only0),
                  ic.ICMElement("adjust_to_minute", integers.Only0),
                  ic.ICMElement("adjust_to_preset_time", integers.Only0),
                  ic.ICMElement("preset_adjusting_time", PresetAdjustingTime),
                  ic.ICMElement("shift_time", ShiftTime))

    def characteristics_init(self):
        self.cardinality = (0, 1)

    @property
    def time(self) -> cst.OctetStringDateTime:
        return self.get_attr(2)

    @property
    def time_zone(self) -> cdt.Long:
        return self.get_attr(3)

    @property
    def status(self) -> ClockStatus:
        return self.get_attr(4)

    @property
    def daylight_savings_begin(self) -> cst.OctetStringDateTime:
        return self.get_attr(5)

    @property
    def daylight_savings_end(self) -> cst.OctetStringDateTime:
        return self.get_attr(6)

    @property
    def daylight_savings_deviation(self) -> DaylightSavingsDeviation:
        return self.get_attr(7)

    @property
    def daylight_savings_enabled(self) -> cdt.Boolean:
        return self.get_attr(8)

    @property
    def clock_base(self) -> ClockBase:
        return self.get_attr(9)

    @property
    def adjust_to_quarter(self) -> integers.Only0:
        return self.get_meth(1)

    @property
    def adjust_to_measuring_period(self) -> integers.Only0:
        return self.get_meth(2)

    @property
    def adjust_to_minute(self) -> integers.Only0:
        return self.get_meth(3)

    @property
    def adjust_to_preset_time(self) -> integers.Only0:
        return self.get_meth(4)

    @property
    def preset_adjusting_time(self) -> PresetAdjustingTime:
        return self.get_meth(5)

    @property
    def shift_time(self) -> ShiftTime:
        return self.get_meth(6)
